import json
import asyncio

from .service import Service, ServiceResponse

'''
# Discovery Broadcast
{
    'services': ['text']
}
'''

class DiscoveryProtocol:
    def start_discovery(self):
        if hasattr(self, 'task') and self.task is not None:
            self.task.cancel()
        self.task = asyncio.Task(self.discovery_tick())

    async def discovery_tick(self):
        while True:
            services = []
            for id, service in self.epyphany.services.items():
                data = {'id': id, 'port': service.port, 'payload': service.create_payload()}
                if data['payload'] is not False:
                    services.append(data)
            self.broadcast({'services': services})
            await asyncio.sleep(5)

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, data, addr):
        message = json.loads(data.decode())

        if 'services' in message:
            services = []
            for data in message['services']:
                data['ip'] = addr[0]
                service = ServiceResponse(**data)
                services.append(service)
            self.epyphany.on_discovery(self, services)

    def broadcast(self, data):
        self.transport.sendto(json.dumps(data).encode('UTF-8'), ('<broadcast>', self.epyphany.port))

    def send(self, addr, data):
        self.transport.sendto(json.dumps(data).encode('UTF-8'), addr)



import logging
import asyncio
import socket

from .protocol import DiscoveryProtocol
from .service import Service


class Epyphany:
    def __init__(self, port, loop=asyncio.get_event_loop()):
        self.logger = logging.getLogger('Epyphany')

        self.port = port

        self.services = {}
        self.listeners = {'*': []}

        self.loop = loop
        self.listen_task = None
        self.protocol = None

    def on_discovery(self, protocol, services):
        for service in services:
            if service.id in self.listeners:
                for listener in self.listeners[service.id]:
                    listener(protocol, service)

            for listener in self.listeners['*']:
                listener(protocol, service)

    def register_service(self, service):
        self.logger.debug('Registering service: %r' % service.id)

        self.services[service.id] = service
        service.loop = self.loop
        service.start_service()

    def register_listener(self, arg):
        if callable(arg):
            self.logger.debug('Registered listener for \'*\': %r' % arg.__name__)

            self.listeners['*'].append(arg)
        else:
            arg = arg.upper()

            if arg not in self.listeners:
                self.listeners[arg] = []

            def inner(func):
                self.logger.debug('Registered listener for %r: %r' % (arg, func.__name__))

                self.listeners[arg].append(func)
                return func
            return inner

    def begin(self):
        self.start()
        self.discover()

    def start(self):
        self.logger.info('Initializing discovery...')

        if self.listen_task is not None: return

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.bind(('', self.port))

        self.listen_task = self.loop.create_datagram_endpoint(DiscoveryProtocol, sock=self._sock)
        self.protocol = self.loop.run_until_complete(self.listen_task)[1]
        self.protocol.epyphany = self

    def discover(self):
        self.protocol.start_discovery()

    def run_forever(self):
        self.loop.run_forever()



import os
from distutils.core import setup


setup(
    name = 'epyphany',
    version = '0.1.2',
    packages=['epyphany'],
    description = ('A script for facilitating network service discovery.'),
    python_requires='>=3.5.3',

    author = 'Stumblinbear',
    author_email = 'stumblinbear@gmail.com',

    license = 'GNU',
    keywords = 'network discovery discover service broadcast packet',
    url = 'https://github.com/Secret-Web/epyphany',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Topic :: System :: Networking'
    ],
)



class Service:
    def __init__(self, id, port):
        self._id = id.upper()
        self._port = port

    @property
    def id(self):
        return self._id

    @property
    def port(self):
        return self._port

    def start_service(self):
        pass

    def create_payload(self):
        pass

class ServiceResponse:
    def __init__(self, id, ip, port, payload={}):
        self.id = id
        self.ip = ip
        self.port = port
        self.payload = payload

    def __hash__(self):
        return hash((self.id, self.ip, self.port))

    def __eq__(self, cmp):
        return hash(self) == hash(cmp)

    def __repr__(self):
        return '%s <%s:%d>' % (self.id, self.ip, self.port)
